/**
 * Used to specify the strength of a CoveringArray
 */
type CoveringStrength = 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18 | 19;
/**
 * Type used as an index within recursive iterations
 */
type Prev = [
    never,
    0,
    1,
    2,
    3,
    4,
    5,
    6,
    7,
    8,
    9,
    10,
    11,
    12,
    13,
    14,
    15,
    16,
    17,
    18,
    19,
    20,
    ...0[]
];
/**
 * Concatenates two strings into one
 */
type Join<K, P> = K extends string | number ? P extends string | number ? `${K}${"" extends P ? "" : "."}${P}` : never : never;
/**
 * Path to a leaf element in an object (keys concatenated by dots)
 * If D extends number = 10,
 * Error: Type instantiation is excessively deep and possibly infinite.(2589)
 * Thus, it's reduced to 4.
 * // 例
 * type Example = {
 * 	a:{
 * 		x: [
 * 			"ppp",
 * 			"qqq",
 * 		],
 * 		y: [
 *
 * 		],
 * 	},
 * 	b: number,
 * }
 *
 * Path<Example> // 'a.x'|'a,y'|'b'
 */
export type Path<T, D extends number = 4> = [
    D
] extends [never] ? never : T extends readonly string[] ? "" : {
    [K in keyof T]-?: Join<K, Path<T[K], Prev[D]>>;
}[keyof T];
/**
 * Type of the leaf element in an object
 * // Example
 * type Example = {
 * 	a:{
 * 		x: number,
 * 		y: number,
 * 	},
 * 	b: 'あ' | 'い' | 'う',
 * }
 *
 * Leaf<Example, 'a.x'> // number
 * Leaf<Example, 'b'> 	// 'あ' | 'い' | 'う'
 */
type Leaf<T, PATH extends string> = PATH extends Path<T> ? PATH extends `${infer KEY}.${infer REST}` ? KEY extends keyof T ? Leaf<T[KEY], REST> : never : PATH extends keyof T ? T[PATH] : never : never;
/**
 * Represents the range and domain of a tree-structured object
 */
export type TreeDomain = {
    [key: string]: TreeDomain | readonly string[];
};
/**
 * Represents specific values of a tree-structured object
 */
export type Tree<D extends TreeDomain> = {
    [KEY in Extract<keyof D, string>]: D[KEY] extends TreeDomain ? Tree<D[KEY]> : D[KEY][number];
};
/**
 * Flattens an object of tree structure to a one-dimensional associative array.
 * Keys are concatenated with dots like `Given.User.Type`.
 */
export type ListDomain<D extends TreeDomain> = {
    [KEY in Path<D>]?: Leaf<D, KEY>;
};
/**
 * Flattens an object of tree structure to a one-dimensional associative array.
 * Keys are concatenated with dots like `Given.User.Type`.
 * Keys are not required.
 */
export type List<D extends TreeDomain> = {
    [KEY in ("ID" | "Perspective" | Path<D>)]?: "ID" extends KEY ? string : "Perspective" extends KEY ? string : Leaf<D, KEY> extends readonly string[] ? Leaf<D, KEY>[number] : never;
};
/**
 * Default values for variables in a testcase
 */
export type Defaults<D extends TreeDomain> = Tree<D>;
/**
 * Testcase
 */
export type Test<D extends TreeDomain> = List<D>;
/**
 * Perspective to expand the testcase
 */
type Perspective<D extends TreeDomain> = {
    readonly "title": string;
    readonly "constants": {
        [KEY in Path<D>]?: Leaf<D, KEY> extends readonly string[] ? Leaf<D, KEY>[number] : never;
    };
    readonly "variables": readonly [Path<D>][number][];
    readonly "expect": (testcase: List<D>) => List<D>;
    readonly "strength"?: CoveringStrength;
};
/**
 * Collection of perspectives for expanding testcases
 */
export type Perspectives<D extends TreeDomain> = readonly Perspective<D>[];
/**
 * Conditions to exclude certain combinations in a testcase
 * If these conditions are met, the testcase is excluded.
 */
type Exclusion<D extends TreeDomain> = {
    readonly [KEY in Path<D>]?: Leaf<D, KEY> extends readonly string[] ? Leaf<D, KEY>[number] | RegExp : never;
};
/**
 * Collection of conditions to exclude certain combinations in a testcase
 */
export type Exclusions<D extends TreeDomain> = readonly Exclusion<D>[];
/**
 * Options for file output
 */
export type FileOutputOption = {
    json?: {
        file_path: string;
    };
    markdown?: {
        file_path: string;
        true_symbol?: string;
        false_symbol?: string;
    };
};
/**
 * Generates a complete set of testcases.
 * @param domain
 * @param defaults
 * @param exclusions
 * @param perspectives
 * @returns
 */
export declare function generateTests<D extends TreeDomain>(domain: D, defaults: Defaults<D>, exclusions: Exclusions<D>, perspectives: Perspectives<D>, file_out_option?: FileOutputOption): Test<D>[];
export {};
