"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateTests = void 0;
//=============================================================================================================
//	Functions
//=============================================================================================================
/**
 * Generates a complete set of testcases.
 * @param domain
 * @param defaults
 * @param exclusions
 * @param perspectives
 * @returns
 */
function generateTests(domain, defaults, exclusions, perspectives, file_out_option) {
    // Retrieve all parameters (example: "Given.User.Type", "When.Browser.Start")
    let list_domain = {};
    treeDomainToListDomain(list_domain, domain);
    // Create the default testcase
    let list_default = {};
    treeToList(list_default, defaults);
    // Create the list of tests
    let tests = [];
    for (let perspective of perspectives) {
        //------------------------------------------------
        // Expand based on a particular perspective
        //------------------------------------------------
        //------------------------------------------------
        // Reflect constants in the base testcase
        //------------------------------------------------
        let base_test = Object.assign({}, list_default);
        for (let key in perspective.constants) {
            base_test[key] = perspective.constants[key];
        }
        //------------------------------------------------
        // Expand variables and reflect them in the testcase
        //------------------------------------------------
        tests = [
            ...tests,
            ...generateTestsByPerspective(list_domain, perspective, base_test, exclusions)
        ];
    }
    // Assign ID to all testcases
    let id = 1;
    for (let t in tests) {
        tests[t].ID = `${id}`;
        id++;
    }
    //-----------------------------------------------------
    // File output
    //-----------------------------------------------------
    if (file_out_option) {
        const fs = require('fs');
        //-----------------------
        // JSON format
        //-----------------------
        if (file_out_option.json) {
            fs.writeFile(`${file_out_option.json.file_path}`, JSON.stringify(tests, null, '\t'), (err) => { if (err)
                console.log(err); });
        }
        //-----------------------
        // Markdown format
        //-----------------------
        if (file_out_option.markdown) {
            // Convert to a table
            fs.writeFile(`${file_out_option.markdown.file_path}`, toMarkdown(toTable(list_domain, tests, file_out_option.markdown.true_symbol, file_out_option.markdown.false_symbol)), (err) => { if (err)
                console.log(err); });
        }
    }
    return tests;
}
exports.generateTests = generateTests;
/**
 * Converts variables from TreeDomain type to ListDomain type.
 * The data structure is converted from a tree to a one-dimensional associative array,
 * but the structure of the endpoints (leaves) remains unchanged (array of Strings).
 * @param result
 * @param tree
 * @param parent_path
 */
function treeDomainToListDomain(result, tree, parent_path = '') {
    for (const key in tree) {
        const value = tree[key];
        const path = (parent_path === '' ? key : parent_path + '.' + key);
        if (isLeafOfTreeDomain(value)) {
            // It's a leaf node, so register the choices
            if (typeof result[path] !== "undefined") {
                throw Error('Duplicate definitions exist.');
            }
            result[path] = value;
        }
        else {
            // Further recursively expand
            treeDomainToListDomain(result, value, path);
        }
    }
}
/**
 * Converts variables from Tree type to List type.
 * The data structure is converted from a tree to a one-dimensional associative array,
 * but the structure of the endpoints (leaves) remains unchanged (String).
 * @param result
 * @param tree
 * @param parent_path
 */
function treeToList(result, tree, parent_path = '') {
    for (const key in tree) {
        const value = tree[key];
        const path = (parent_path === '' ? key : parent_path + '.' + key);
        if (isLeafOfTree(value)) {
            // It's a leaf node, so register the value
            if (typeof result[path] !== "undefined") {
                throw Error('Duplicate definitions exist.');
            }
            result[path] = value;
        }
        else {
            // Further recursively expand
            treeToList(result, value, path);
        }
    }
}
/**
 * Generates test cases based on a perspective.
 * A recursive function.
 * @param tests
 * @param list_domain
 * @param perspective
 * @param base_test
 * @param exclusions
 * @param v
 */
function generateTestsByPerspective(list_domain, perspective, base_test, exclusions) {
    //---------------------------------------------------
    // 1. Retrieve the complete set of required combinations
    //---------------------------------------------------
    let requirements = getRequirements(list_domain, perspective, exclusions);
    //---------------------------------------------------
    // 2. Generate a Covering Array
    //---------------------------------------------------
    let tests = [];
    generateCoveringArray(tests, requirements);
    //---------------------------------------------------
    // 3. Fill empty cells with default values
    //---------------------------------------------------
    applyBaseTestToCoveringArray(tests, base_test);
    //---------------------------------------------------
    // 4. Exclude tests that match prohibitions
    //---------------------------------------------------
    tests = tests.filter(test => {
        return !isExcluded(test, exclusions);
    });
    //---------------------------------------------------
    // 5. Add Perspective and Expected Results
    //---------------------------------------------------
    for (const t in tests) {
        // Add perspective name
        tests[t]["Perspective"] = perspective.title.replace(/\(/gi, '（').replace(/\)/gi, '）');
        // Add expected results
        perspective.expect(tests[t]);
    }
    return tests;
}
function getRequirements(list_domain, perspective, exclusions) {
    var _a;
    // Runtime input check
    if (perspective.strength > perspective.variables.length) {
        throw new Error("Coverage strength must be less than the length of variables.");
    }
    // Determine strength
    const strength = (_a = perspective.strength) !== null && _a !== void 0 ? _a : perspective.variables.length;
    const variables_combinations = getSubsetArrays(perspective.variables, strength);
    let requirements = [];
    for (const combination of variables_combinations) {
        getFullCoverage(requirements, {}, list_domain, combination, exclusions, 0);
    }
    return requirements;
}
/**
 * Based on the input array, retrieves subsets composed of specified element counts
 * @param array
 * @param this_index
 * @param num_to_choose
 * @returns
 */
function getSubsetArrays(array, m) {
    let result = [];
    if (array.length < m) {
        return [];
    }
    if (m === 1) {
        for (let i = 0; i < array.length; i++) {
            result[i] = [array[i]];
        }
    }
    else {
        for (let i = 0; i < array.length - m + 1; i++) {
            let row = getSubsetArrays(array.slice(i + 1), m - 1);
            for (let j = 0; j < row.length; j++) {
                result.push([array[i]].concat(row[j]));
            }
        }
    }
    return result;
}
/**
 * Retrieves a complete coverage list concerning variables
 * @param lists
 * @param base_list
 * @param list_domain
 * @param variables
 * @param exclusions
 * @param v
 */
function getFullCoverage(lists, base_list, list_domain, variables, exclusions, v) {
    let list = Object.assign({}, base_list);
    if (variables.length > v) {
        //-------------------------------------------------
        // Recur until the required combination is fully covered
        //-------------------------------------------------
        const variable = variables[v];
        for (let c in list_domain[variable]) {
            const choice = list_domain[variable][c];
            list[variable] = choice;
            getFullCoverage(lists, list, list_domain, variables, exclusions, v + 1);
        }
    }
    else {
        //-------------------------------------------------
        // Once the required combination is fully covered, add it
        //-------------------------------------------------
        // Push only if not excluded
        if (!isExcluded(list, exclusions)) {
            lists.push(list);
        }
    }
}
/**
 * Scans the required combinations and generates a Covering Array
 * @param tests
 * @param requirements
 */
function generateCoveringArray(tests, requirements) {
    for (const req of requirements) {
        // Check for existence
        const keys = Object.keys(req);
        const exists = tests.some(test => {
            // Check for rows where all factors of the combination match
            return keys.every(key => test[key] === req[key]);
        });
        if (exists) {
            //---------------------------------------------------------
            // If there is a row that meets req, proceed to the next req
            //---------------------------------------------------------
            continue;
        }
        else {
            //---------------------------------------------------------
            // Check if there is a row where req can be satisfied by appending cells
            //---------------------------------------------------------
            const appendables = tests
                .map((test, t) => {
                return Object.assign({ index: t }, test);
            })
                .filter(test => {
                // Check for available factors
                return keys.some(key => test[key] === undefined);
            })
                .filter(test => {
                // Check that there are no factors that do not match the required ones
                return !keys.some(key => (test[key] && test[key] !== req[key]));
            });
            if (appendables.length > 0) {
                // Append cells
                const index_to_append = appendables[0].index;
                for (const key of keys) {
                    if (!tests[index_to_append][key]) {
                        tests[index_to_append][key] = req[key];
                    }
                }
            }
            else {
                // Add a row
                let new_test = {};
                for (const key of keys) {
                    new_test[key] = req[key];
                }
                tests.push(new_test);
            }
        }
    }
}
/**
 * Fills empty cells in a Covering Array with default values
 */
function applyBaseTestToCoveringArray(tests, base_test) {
    const keys = Object.keys(base_test);
    for (const t in tests) {
        for (const key of keys) {
            if (!tests[t][key]) {
                // If the cell is empty, fill it with the default value
                tests[t][key] = base_test[key];
            }
        }
    }
}
/**
 * Checks if a testcase matches conditions for exclusion
 * @param test
 * @param exclusions
 * @returns
 */
function isExcluded(test, exclusions) {
    return exclusions.some((exclusion) => {
        // Check if there is at least one matching exclusion condition
        return Object.keys(exclusion).every((key) => {
            // Check if all conditions within the exclusion match
            if (exclusion[key] instanceof RegExp) {
                // Check if the regular expression matches
                return (exclusion[key]).test(test[key]);
            }
            else {
                // Check if the strings match
                return test[key] === exclusion[key];
            }
        });
    });
}
/**
 * Type guard for the endpoint of TreeDomain
 * @param a
 */
function isLeafOfTreeDomain(a) {
    return Array.isArray(a) || typeof a === 'string';
}
/**
 * Type guard for the endpoint of a Tree
 * @param a
 */
function isLeafOfTree(a) {
    return !Array.isArray(a) && typeof a === 'string';
}
/**
 * Converts a complete set of test cases into a two-dimensional array
 * @param tests
 */
function toTable(test_domain, tests, true_symbol = 'T', false_symbol = 'F') {
    const keys = Object.keys(test_domain);
    let table = [];
    //------------------------------------
    // Header part
    //------------------------------------
    for (const key of keys) {
        for (const value of test_domain[key]) {
            table.push([key, value]);
        }
    }
    //------------------------------------
    // Body part
    //------------------------------------
    for (const row of table) {
        const key = row[0];
        const val = row[1];
        for (const test of tests) {
            row.push(test[key] === val ? true_symbol : false_symbol);
        }
    }
    return table;
}
/**
 * Converts a two-dimensional array into a markdown table
 * @param tests
 */
function toMarkdown(table) {
    let result = '';
    if (table.length > 0) {
        const rows_num = table.length;
        const cols_num = table[0].length;
        const tests_num = cols_num - 2;
        //--------------------------
        // Header part
        //--------------------------
        result += `|||`;
        for (let t = 0; t < tests_num; t++) {
            result += `#${t + 1}|`;
        }
        result += '\r\n';
        result += '|';
        result += '--|'.repeat(cols_num);
        result += '\r\n';
        //--------------------------
        // Body part
        //--------------------------
        let written_keys = [];
        for (let r = 0; r < rows_num; r++) {
            result += '|';
            for (let c = 0; c < cols_num; c++) {
                if (c === 0) {
                    //----------------------------------------------------------
                    // Only factor columns, processing to ensure the same value is not written out twice
                    //----------------------------------------------------------
                    if (written_keys.includes(table[r][c])) {
                        // If the factor has already been written out,
                        // write out nothing
                        result += `|`;
                    }
                    else {
                        // If the factor has not been written out,
                        // record that it has been written
                        written_keys.push(table[r][c]);
                        result += `${table[r][c]}|`;
                    }
                }
                else {
                    result += `${table[r][c]}|`;
                }
            }
            result += '\r\n';
        }
    }
    return result;
}
